@@ -1,32 +1,22 @@
-# -*- coding: utf-8 -*-
-
-from __future__ import unicode_literals
-
 import logging
 
+from skosprovider.skos import label as skoslabel
+from sqlalchemy import Boolean
+from sqlalchemy import Column
+from sqlalchemy import ForeignKey
+from sqlalchemy import Integer
+from sqlalchemy import String
+from sqlalchemy import Table
+from sqlalchemy import Text
+from sqlalchemy import UniqueConstraint
+from sqlalchemy import event
+from sqlalchemy import orm
+from sqlalchemy.ext.hybrid import hybrid_property
+from sqlalchemy.orm import backref
+from sqlalchemy.orm import relationship
+
 log = logging.getLogger(__name__)
-
-from language_tags import tags
-
-from sqlalchemy import (
-    Column,
-    Integer,
-    Text,
-    String,
-    ForeignKey,
-    UniqueConstraint,
-    Table,
-    event
-)
-
-from sqlalchemy.ext.declarative import declarative_base
-
-from sqlalchemy.orm import (
-    relationship,
-    backref
-)
-
-Base = declarative_base()
+Base = orm.declarative_base()
 
 concept_label = Table(
     'concept_label',
@@ -54,6 +44,13 @@ concept_note = Table(
     Column('note_id', Integer, ForeignKey('note.id'), primary_key=True)
 )
 
+concept_source = Table(
+    'concept_source',
+    Base.metadata,
+    Column('concept_id', Integer, ForeignKey('concept.id'), primary_key=True),
+    Column('source_id', Integer, ForeignKey('source.id'), primary_key=True)
+)
+
 conceptscheme_note = Table(
     'conceptscheme_note',
     Base.metadata,
@@ -66,6 +63,40 @@ conceptscheme_note = Table(
     Column('note_id', Integer, ForeignKey('note.id'), primary_key=True)
 )
 
+conceptscheme_source = Table(
+    'conceptscheme_source',
+    Base.metadata,
+    Column(
+        'conceptscheme_id',
+        Integer,
+        ForeignKey('conceptscheme.id'),
+        primary_key=True
+    ),
+    Column(
+        'source_id',
+        Integer,
+        ForeignKey('source.id'),
+        primary_key=True
+    )
+)
+
+conceptscheme_language = Table(
+    'conceptscheme_language',
+    Base.metadata,
+    Column(
+        'conceptscheme_id',
+        Integer,
+        ForeignKey('conceptscheme.id'),
+        primary_key=True
+    ),
+    Column(
+        'language_id',
+        String(64),
+        ForeignKey('language.id'),
+        primary_key=True
+    )
+)
+
 collection_concept = Table(
     'collection_concept',
     Base.metadata,
@@ -141,7 +172,7 @@ class Thing(Base):
     id = Column(Integer, primary_key=True)
     type = Column(String(30))
     concept_id = Column(
-        Integer,
+        String,
         nullable=False,
         index=True
     )
@@ -160,8 +191,17 @@ class Thing(Base):
         cascade='all, delete-orphan',
         single_parent=True
     )
+    sources = relationship(
+        'Source',
+        secondary=concept_source,
+        backref=backref('concept', uselist=False),
+        cascade='all, delete-orphan',
+        single_parent=True
+    )
 
-    conceptscheme = relationship('ConceptScheme', backref='concepts')
+    conceptscheme = relationship(
+        'ConceptScheme', backref=orm.backref('concepts', cascade_backrefs=False)
+    )
     conceptscheme_id = Column(
         Integer,
         ForeignKey('conceptscheme.id'),
@@ -171,13 +211,33 @@ class Thing(Base):
 
 
     def label(self, language='any'):
-        return label(self.labels, language)
+        return skoslabel(self.labels, language)
+
+    def _sortkey(self, key='id', language='any'):
+        '''
+        Provide a single sortkey.
+
+        :param string key: Either `id`, `uri`, `label` or `sortlabel`.
+        :param string language: The preferred language to receive the label in
+            if key is `label` or `sortlabel`. This should be a valid IANA language tag.
+        :rtype: :class:`str`
+        '''
+        if key == 'id':
+            return str(self.id)
+        elif key == 'uri':
+            return self.uri if self.uri else ''
+        else:
+            l = skoslabel(self.labels, language, key == 'sortlabel')
+            return l.label.lower() if l else ''
 
     __mapper_args__ = {
         'polymorphic_on': 'type',
         'polymorphic_identity': 'thing'
     }
 
+    def __str__(self):
+        return self.__class__.__name__ + '-' + str(self.id)
+
 
 class Concept(Thing):
     '''
@@ -195,19 +255,21 @@ class Concept(Thing):
     narrower_concepts = relationship(
         'Concept',
         secondary=concept_hierarchy_concept,
-        backref=backref('broader_concepts', collection_class=set),
+        backref=backref('broader_concepts', collection_class=set, cascade_backrefs=False),
         primaryjoin='Concept.id==concept_hierarchy_concept.c.concept_id_broader',
         secondaryjoin='Concept.id==concept_hierarchy_concept.c.concept_id_narrower',
-        collection_class=set
+        collection_class=set,
+        cascade_backrefs=False,
     )
 
     narrower_collections = relationship(
         'Collection',
         secondary=concept_hierarchy_collection,
-        backref=backref('broader_concepts', collection_class=set),
+        backref=backref('broader_concepts', collection_class=set, cascade_backrefs=False),
         primaryjoin='Concept.id==concept_hierarchy_collection.c.concept_id_broader',
         secondaryjoin='Concept.id==concept_hierarchy_collection.c.collection_id_narrower',
-        collection_class=set
+        collection_class=set,
+        cascade_backrefs=False,
     )
 
     __mapper_args__ = {
@@ -262,19 +324,21 @@ class Collection(Thing):
     A collection as know by :term:`skosprovider:SKOS`.
     '''
 
-    __mapper_args__ = {
-        'polymorphic_identity': 'collection'
-    }
+    infer_concept_relations = Column(Boolean, nullable=False, default=True)
 
     members = relationship(
         'Thing',
         secondary=collection_concept,
-        backref=backref('member_of', collection_class=set),
+        backref=backref('member_of', collection_class=set, cascade_backrefs=False),
         primaryjoin='Thing.id==collection_concept.c.collection_id',
         secondaryjoin='Thing.id==collection_concept.c.concept_id',
         collection_class=set
     )
 
+    __mapper_args__ = {
+        'polymorphic_identity': 'collection'
+    }
+
 
 class ConceptScheme(Base):
     '''
@@ -282,7 +346,7 @@ class ConceptScheme(Base):
     '''
     __tablename__ = 'conceptscheme'
     id = Column(Integer, primary_key=True)
-    uri = Column(String(512))
+    uri = Column(String(512), nullable=False)
 
     labels = relationship(
         'Label',
@@ -294,9 +358,22 @@ class ConceptScheme(Base):
         secondary=conceptscheme_note,
         backref=backref('conceptscheme', uselist=False)
     )
-
+    languages = relationship(
+        'Language',
+        secondary=conceptscheme_language
+    )
+    sources = relationship(
+        'Source',
+        secondary=conceptscheme_source,
+        backref=backref('conceptscheme', uselist=False),
+        cascade='all, delete-orphan',
+        single_parent=True
+    )
     def label(self, language='any'):
-        return label(self.labels, language)
+        return skoslabel(self.labels, language)
+
+    def __str__(self):
+        return self.__class__.__name__ + '-' + str(self.id)
 
 
 class Language(Base):
@@ -312,7 +389,7 @@ class Language(Base):
         self.name = name
 
     def __str__(self):
-        return self.name
+        return self.id
 
 
 class LabelType(Base):
@@ -344,7 +421,7 @@ class Label(Base):
     )
 
     labeltype = relationship('LabelType', uselist=False)
-    language = relationship('Language', uselist=False)
+    _language = relationship('Language', uselist=False)
 
     labeltype_id = Column(
         String(20),
@@ -353,7 +430,7 @@ class Label(Base):
         index=True
     )
     language_id = Column(
-        String(10),
+        String(64),
         ForeignKey('language.id'),
         nullable=True,
         index=True
@@ -364,6 +441,14 @@ class Label(Base):
         self.language_id = language_id
         self.label = label
 
+    @hybrid_property
+    def type(self):
+        return self.labeltype_id
+
+    @hybrid_property
+    def language(self):
+        return self._language or self.language_id
+
     def __str__(self):
         return self.label
 
@@ -407,21 +492,43 @@ class Note(Base):
 
     language = relationship('Language', uselist=False)
     language_id = Column(
-        String(10),
+        String(64),
         ForeignKey('language.id'),
         nullable=True,
         index=True
     )
+    markup = Column(String(20), nullable=True)
 
-    def __init__(self, note, notetype_id, language_id):
+    def __init__(self, note, notetype_id, language_id, markup=None):
         self.notetype_id = notetype_id
         self.language_id = language_id
         self.note = note
+        self.markup = markup
 
     def __str__(self):
         return self.note
 
 
+class Source(Base):
+    '''
+    The source where a certain piece of information came from.
+    '''
+    __tablename__ = 'source'
+    id = Column(Integer, primary_key=True)
+    citation = Column(
+        Text,
+        nullable=False
+    )
+    markup = Column(String(20), nullable=True)
+
+    def __init__(self, citation, markup=None):
+        self.citation = citation
+        self.markup = markup
+
+    def __str__(self):
+        return self.citation
+
+
 class MatchType(Base):
     '''
     A matchType according to :term:`skosprovider:SKOS`.
@@ -441,7 +548,7 @@ class MatchType(Base):
 
 class Match(Base):
     '''
-    A match between a :class:`Concept` in one ConceptScheme and those in 
+    A match between a :class:`Concept` in one ConceptScheme and those in
     another one.
     '''
     __tablename__ = 'match'
@@ -502,69 +609,32 @@ class Visitation(Base):
         index=True
     )
 
+    def __str__(self):
+        return self.__class__.__name__ + '-' + str(self.id)
 
-def label(labels=[], language='any'):
+
+def label(labels=[], language='any', sortLabel=False):
     '''
     Provide a label for a list of labels.
 
-    The items in the list of labels are assumed to be instances of
-    :class:`Label`.
-
-    This method tries to find a label by looking if there's
-    a pref label for the specified language. If there's no pref label,
-    it looks for an alt label. It disregards hidden labels.
-
-    While matching languages, preference will be given to exact matches. But,
-    if no exact match is present, an inexact match will be attempted. This might
-    be because a label in language `nl-BE` is being requested, but only `nl` or
-    even `nl-NL` is present. Similarly, when requesting `nl`, a label with
-    language `nl-NL` or even `nl-Latn-NL` will also be considered, 
-    providing no label is present that has an exact match with the 
-    requested language.
-
-    If language 'any' was specified, all labels will be considered,
-    regardless of language.
-
-    To find a label without a specified language, pass `None` as language.
-
-    If a language or None was specified, and no label could be found, this
-    method will automatically try to find a label in some other language.
-
-    Finally, if no label could be found, None is returned.
+    .. deprecated:: 0.5.0
+       Please use :func:`skosprovider.skos.label`. Starting with
+       `skosprovider 0.6.0`, the function can function on
+       :class:`skosprovider_sqlalchemy.models.Label` instances as well.
 
     :param list labels: A list of :class:`labels <Label>`.
-    :param str language: The language for which a label should preferentially 
+    :param str language: The language for which a label should preferentially
         be returned. This should be a valid IANA language tag.
+    :param boolean sortLabel: Should sortLabels be considered or not? If True,
+        sortLabels will be preferred over prefLabels. Bear in mind that these
+        are still language dependent. So, it's possible to have a different
+        sortLabel per language.
     :rtype: A :class:`Label` or `None` if no label could be found.
     '''
-    # Normalise the tag
-    broader_language_tag = None
-    if language != 'any':
-        language = tags.tag(language).format
-        broader_language_tag = tags.tag(language).language
-    pref = None
-    alt = None
-    for l in labels:
-        labeltype = l.labeltype_id or l.labeltype.name
-        if language == 'any' or l.language_id == language:
-            if labeltype == 'prefLabel' and (pref is None or pref.language_id != language):
-                pref = l
-            if labeltype == 'altLabel' and (alt is None or alt.language_id != language):
-                alt = l
-        if broader_language_tag and tags.tag(l.language_id).language and tags.tag(
-                l.language_id).language.format == broader_language_tag.format:
-            if labeltype == 'prefLabel' and pref is None:
-                pref = l
-            if labeltype == 'altLabel' and alt is None:
-                alt = l
-    if pref is not None:
-        return pref
-    elif alt is not None:
-        return alt
-    return label(labels, 'any') if language != 'any' else None
+    return skoslabel(labels, language, sortLabel)
 
 
-class Initialiser(object):
+class Initialiser:
     '''
     Initialises a database.
 
@@ -611,7 +681,8 @@ class Initialiser(object):
         labeltypes = [
             ('hiddenLabel', 'A hidden label.'),
             ('altLabel', 'An alternative label.'),
-            ('prefLabel', 'A preferred label.')
+            ('prefLabel', 'A preferred label.'),
+            ('sortLabel', 'A label exclusively used for sorting.')
         ]
         for l in labeltypes:
             lt = LabelType(l[0], l[1])
